from signalbot import Command, regex_triggered
from signalbot import Context as ChatContext

from signalblast.broadcastbot import BroadcasBot
from signalblast.commands_strings import AdminCommandStrings, CommandRegex


class AddAdmin(Command):
    def __init__(self, bot: BroadcasBot) -> None:
        super().__init__()
        self.broadcastbot = bot

    @regex_triggered(CommandRegex.add_admin)
    async def handle(self, ctx: ChatContext) -> None:
        try:
            await ctx.receipt(receipt_type="read")

            subscriber_uuid = ctx.message.source_uuid
            password = self.broadcastbot.message_handler.remove_command_from_message(
                ctx.message.text,
                AdminCommandStrings.add_admin,
            )

            previous_admin = self.broadcastbot.admin.admin_id
            if await self.broadcastbot.admin.add(subscriber_uuid, password):
                await self.broadcastbot.reply_with_warn_on_failure(ctx, "You have been added as admin!")
                if previous_admin is not None and subscriber_uuid != previous_admin:
                    msg_to_admin = self.broadcastbot.message_handler.compose_message_to_admin(
                        "You are no longer an admin!",
                        subscriber_uuid,
                    )
                    await self.broadcastbot.send(previous_admin, msg_to_admin)
                log_message = f"Previous admin was {previous_admin}, new admin is {subscriber_uuid}"
                self.broadcastbot.logger.info(log_message)
            else:
                await self.broadcastbot.reply_with_warn_on_failure(ctx, "Adding failed, admin password is incorrect!")
                if previous_admin is not None:
                    msg_to_admin = self.broadcastbot.message_handler.compose_message_to_admin(
                        "Tried to be added as admin",
                        subscriber_uuid,
                    )
                    await self.broadcastbot.send(previous_admin, msg_to_admin)
                self.broadcastbot.logger.warning("%s failed password check for add_admin", subscriber_uuid)
        except Exception:
            self.broadcastbot.logger.exception("")



import asyncio

from signalblast.broadcastbot import BroadcasBot


async def health_check(bot: BroadcasBot, receiver: str, port: int) -> None:
    async def handle_health_check_request(
        _reader: asyncio.streams.StreamReader,
        writer: asyncio.streams.StreamWriter,
    ) -> None:
        bot.logger.info("Handle health check request")

        try:
            # The health check is to send a ping message to receiver
            await asyncio.wait_for(bot.send(receiver, "Ping"), timeout=30)
            response = "HTTP/1.0 200 OK\r\n\r\nOK \r\n"

            bot.logger.info("Health check message sent")
        except Exception:
            bot.logger.exception("")
            response = "HTTP/1.0 500 Internal Server Error\r\n\r\nInternal Server Error \r\n"

        writer.write(response.encode("utf8"))
        await writer.drain()
        writer.close()

        bot.logger.info("Health check performed")

    server = await asyncio.start_server(handle_health_check_request, "localhost", port)
    async with server:
        await server.serve_forever()



import os
from logging import WARNING, Formatter, Logger, StreamHandler, getLogger
from logging.handlers import TimedRotatingFileHandler
from pathlib import Path

from pydantic import BaseModel


def create_or_set_logger(name: str | None, logging_level: int = WARNING, log_file: Path | None = None) -> Logger:
    # Log to console or log to file, keeping the log for two weeks, rotate every Monday.
    handler = StreamHandler() if log_file is None else TimedRotatingFileHandler(log_file, when="W0", backupCount=1)

    formatter = Formatter("%(asctime)s %(name)s [%(levelname)s] - %(funcName)s - %(message)s")
    handler.setFormatter(formatter)

    logger = getLogger(name)
    logger.setLevel(logging_level)
    logger.addHandler(handler)
    return logger


def get_data_path() -> Path:
    return Path(os.getenv("SIGNALBLAST_CONFIG_DIR", Path.home() / ".local/share/signalblast"))


class TimestampData(BaseModel):
    timestamp: int
    author: str
    broadcast_timestamps: dict[str, int]  # subscriber uuid, timestamp



import asyncio
import contextlib
import random
from collections import defaultdict
from re import Pattern

from signalbot import Command, MessageType
from signalbot import Context as ChatContext

from signalblast.broadcastbot import BroadcasBot
from signalblast.commands_strings import CommandRegex, PublicCommandStrings
from signalblast.utils import TimestampData


class Broadcast(Command):
    MAX_FAILED_MSGS = 10

    def __init__(self, bot: BroadcasBot) -> None:
        super().__init__()
        self.broadcastbot = bot
        self.subscribers_num_fails: dict[str, int] = defaultdict(lambda: 0)

    def is_valid_command(self, message: str, invalid_command: Pattern) -> bool:
        return any(regex != invalid_command and regex.search(message) is not None for regex in CommandRegex)

    async def check_send_tasks_results(self, send_tasks: list[asyncio.Task | None], action_str: str) -> dict[str, int]:
        timestamp_data = {}
        for send_task, subscriber in zip(send_tasks, self.broadcastbot.subscribers, strict=False):
            if send_task is not None:
                try:
                    timestamp_data[subscriber] = send_task.result()
                    self.subscribers_num_fails.pop(subscriber, None)
                    self.broadcastbot.logger.info("Message successfully %s %s", action_str, subscriber)
                except Exception:
                    self.subscribers_num_fails[subscriber] += 1
                    self.broadcastbot.logger.exception("Message not %s %s", action_str, subscriber)

        subscribers_to_remove = []
        for subscriber, num_fails in self.subscribers_num_fails.items():
            if num_fails >= Broadcast.MAX_FAILED_MSGS:
                subscribers_to_remove.append(subscriber)
                await self.broadcastbot.subscribers.remove(subscriber)

                remove_message = "The bot is having problems sending you messages. "
                remove_message += "You have been removed from the list. "
                remove_message += "Please update signal, remove old linked devices and try subscribing again."
                with contextlib.suppress(Exception):
                    # Most likely will fail to send the message but try anyway
                    await self.broadcastbot.send(subscriber, remove_message)

        for subscriber in subscribers_to_remove:
            del self.subscribers_num_fails[subscriber]

        return timestamp_data

    async def broadcast(self, ctx: ChatContext) -> None:  # noqa: C901, PLR0915, PLR0912 function is too complex
        broadcast_timestamps: dict[str, int] = {}
        num_subscribers = -1
        attachments_deleted = False
        send_tasks_checked = False
        timestamp_data_saved = False
        send_tasks: list[asyncio.Task | None] = []
        action_str, acting_str = "sent to", "sending"

        try:
            subscriber_uuid = ctx.message.source_uuid
            if subscriber_uuid in self.broadcastbot.banned_users:
                await self.broadcastbot.send(subscriber_uuid, "This number is not allowed to send messages")
                self.broadcastbot.logger.info("%s tried to broadcast but they are banned", subscriber_uuid)
                return

            if subscriber_uuid not in self.broadcastbot.subscribers:
                await self.broadcastbot.send(subscriber_uuid, self.broadcastbot.must_subscribe_message)
                self.broadcastbot.logger.info("%s tried to broadcast but they are not subscribed", subscriber_uuid)
                return

            num_subscribers = len(self.broadcastbot.subscribers)

            message = self.broadcastbot.message_handler.remove_command_from_message(
                ctx.message.text,
                PublicCommandStrings.broadcast,
            )
            attachments = self.broadcastbot.message_handler.empty_list_to_none(ctx.message.base64_attachments)
            link_preview = ctx.message.link_previews[0] if len(ctx.message.link_previews) > 0 else None

            if message is None and attachments is None and ctx.message.type != MessageType.DELETE_MESSAGE:
                return

            if message is None:
                message = ""

            # Broadcast message to all subscribers.
            send_tasks: list[asyncio.Task | None] = [None] * num_subscribers

            if ctx.message.type in (MessageType.DELETE_MESSAGE, MessageType.EDIT_MESSAGE):
                if ctx.message.type == MessageType.DELETE_MESSAGE:
                    action_str, acting_str = "deleted for", "deleting"
                    original_msg_timestamp = ctx.message.remote_delete_timestamp
                else:
                    action_str, acting_str = "edited for", "editing"
                    original_msg_timestamp = ctx.message.target_sent_timestamp

                self.broadcastbot.storage_lock.acquire()
                prev_timestamps = ctx.bot.storage.read(
                    f"broadcast-uuid-{subscriber_uuid}-timestamp-{original_msg_timestamp}",
                )
                self.broadcastbot.storage_lock.release()
                to_modify_timestamps = TimestampData.model_validate(prev_timestamps).broadcast_timestamps

            else:
                to_modify_timestamps = {}

            for i, subscriber in enumerate(self.broadcastbot.subscribers):
                if ctx.message.type == MessageType.DELETE_MESSAGE:
                    subscriber_task = ctx.bot.remote_delete(subscriber, to_modify_timestamps.get(subscriber))
                else:
                    subscriber_task = self.broadcastbot.send(
                        subscriber,
                        message,
                        base64_attachments=attachments,
                        link_preview=link_preview,
                        edit_timestamp=to_modify_timestamps.get(subscriber),
                        view_once=ctx.message.view_once,
                    )
                send_tasks[i] = asyncio.create_task(subscriber_task)

                # Avoid rate limiting by waiting a random time between messages
                await asyncio.sleep(random.uniform(0.5, 1))  # noqa: S311

            await asyncio.wait(send_tasks)

            broadcast_timestamps = await self.check_send_tasks_results(send_tasks, action_str)
            send_tasks_checked = True

            if ctx.message.type != MessageType.DELETE_MESSAGE:
                broadcastdata = TimestampData(
                    author=subscriber_uuid,
                    timestamp=ctx.message.timestamp,
                    broadcast_timestamps=broadcast_timestamps,
                )

                self.broadcastbot.storage_lock.acquire()
                ctx.bot.storage.save(
                    f"broadcast-uuid-{subscriber_uuid}-timestamp-{ctx.message.timestamp}",
                    broadcastdata.model_dump(),
                )
                self.broadcastbot.storage_lock.release()
            timestamp_data_saved = True

            await self.broadcastbot.message_handler.delete_attachments(ctx)
            attachments_deleted = True

            await self.broadcastbot.reply_with_warn_on_failure(
                ctx,
                f"Message {action_str} {len(broadcast_timestamps) - 1} people",
            )

            self.broadcastbot.last_msg_user_uuid = subscriber_uuid
        except Exception:
            self.broadcastbot.logger.exception("")
            try:
                if send_tasks_checked is False:
                    broadcast_timestamps = await self.check_send_tasks_results(send_tasks, action_str)

                error_str = f"Something went wrong when {acting_str} the message"
                error_str += (
                    f", it was only {action_str} {len(broadcast_timestamps) - 1} out of {num_subscribers - 1} people"
                )
                error_str += ", please contact the admin if the problem persists"
                await self.broadcastbot.reply_with_warn_on_failure(ctx, error_str)

                if attachments_deleted is False:
                    await self.broadcastbot.message_handler.delete_attachments(ctx)

                if timestamp_data_saved is False and ctx.message.type != MessageType.DELETE_MESSAGE:
                    broadcastdata = TimestampData(
                        author=subscriber_uuid,
                        timestamp=ctx.message.timestamp,
                        broadcast_timestamps=broadcast_timestamps,
                    )
                    self.broadcastbot.storage_lock.acquire()
                    ctx.bot.storage.save(
                        f"broadcast-uuid-{subscriber_uuid}-timestamp-{ctx.message.timestamp}",
                        broadcastdata.model_dump(),
                    )
                    self.broadcastbot.storage_lock.release()
            except Exception:
                self.broadcastbot.logger.exception("")

    async def handle(self, ctx: ChatContext) -> None:
        message = ctx.message.text
        subscriber_uuid = ctx.message.source_uuid

        # Delete a previously broadcasted message
        if ctx.message.type == MessageType.DELETE_MESSAGE:
            await self.broadcast(ctx)
            return

        if message is None:
            if ctx.message.base64_attachments == []:
                self.broadcastbot.logger.info("Received reaction, sticker or similar from %s", subscriber_uuid)
                return

            await ctx.receipt(receipt_type="read")

            # Only attachment, assume the user wants to forward that
            self.broadcastbot.logger.info("Received a file from %s, broadcasting!", subscriber_uuid)
            await self.broadcast(ctx)
            return

        if self.is_valid_command(message, invalid_command=CommandRegex.broadcast):
            return

        await ctx.receipt(receipt_type="read")

        # By default broadcast all the messages
        await self.broadcast(ctx)



from signalbot import Command, regex_triggered
from signalbot import Context as ChatContext

from signalblast.broadcastbot import BroadcasBot
from signalblast.commands_strings import AdminCommandStrings, CommandRegex


class SetPing(Command):
    def __init__(self, bot: BroadcasBot) -> None:
        super().__init__()
        self.broadcastbot = bot

    async def _send_ping(self, receiver: str) -> None:
        try:
            await self.broadcastbot.send(receiver, "Ping")
        except Exception:
            self.broadcastbot.logger.exception("")
            try:
                await self.broadcastbot.send(receiver, "Failed to send ping")
            except Exception:
                self.broadcastbot.logger.exception("")

    async def process_ping_msg(self, ctx: ChatContext) -> None:
        ping_time = self.broadcastbot.message_handler.remove_command_from_message(
            ctx.message.text,
            AdminCommandStrings.set_ping,
        )

        if not await self.broadcastbot.is_user_admin(ctx, AdminCommandStrings.set_ping):
            return

        if ctx.message.group is None:
            error_msg = "Empty group for set ping message"
            raise RuntimeError(error_msg)

        if self.broadcastbot.expiration_time is not None:
            await self.broadcastbot.set_group_expiration_time(ctx.message.group, self.broadcastbot.expiration_time)

        if self.broadcastbot.ping_job is not None:
            self.broadcastbot.scheduler.remove_job(self.broadcastbot.ping_job.id)
            self.broadcastbot.logger.info("Unset old ping job")
            await self.broadcastbot.reply_with_warn_on_failure(ctx, "Unset old ping job")

        self.broadcastbot.ping_job = self.broadcastbot.scheduler.add_job(
            self._send_ping,
            "interval",
            seconds=int(ping_time),
            args=[ctx.message.group],
        )

        await self.broadcastbot.reply_with_warn_on_failure(ctx, f"Ping set every {ping_time} seconds")
        self.broadcastbot.logger.info("Ping set every %s seconds", ping_time)

    @regex_triggered(CommandRegex.set_ping)
    async def handle(self, ctx: ChatContext) -> None:
        try:
            await ctx.receipt(receipt_type="read")

            await self.process_ping_msg(ctx)
        except Exception:
            self.broadcastbot.logger.exception("")
            try:
                await self.broadcastbot.reply_with_warn_on_failure(ctx, "Failed set ping")
            except Exception:
                self.broadcastbot.logger.exception("")



#!/usr/bin/env python

import argparse
import asyncio
import logging
import os

from signalblast.broadcastbot import BroadcasBot
from signalblast.commands import (
    AddAdmin,
    BanSubscriber,
    Broadcast,
    DisplayHelp,
    LastMsgUserUuid,
    LiftBanSubscriber,
    MessageFromAdmin,
    MessageToAdmin,
    RemoveAdmin,
    SetPing,
    Subscribe,
    UnsetPing,
    Unsubscribe,
)
from signalblast.health_check import health_check
from signalblast.log_rollover import rotate_logs_periodically
from signalblast.utils import create_or_set_logger, get_data_path

LOGGING_LEVEL = logging.INFO

# LOG_FILE_PATH = None # Log to console
LOG_FILE_PATH = get_data_path() / "signalblast.log"

create_or_set_logger("signalbot", logging.WARNING, LOG_FILE_PATH)
create_or_set_logger("apscheduler", logging.WARNING, LOG_FILE_PATH)


async def initialise_bot(  # noqa: PLR0913 Too many arguments in function definition
    signal_service: str,
    phone_number: str,
    admin_pass: str,
    expiration_time: int,
    welcome_message: str | None = None,
    health_check_port: int = 15556,
    health_check_receiver: str | None = None,
    instructions_url: str | None = None,
) -> BroadcasBot:
    config = {
        "signal_service": signal_service,
        "phone_number": phone_number,
        "storage": {"type": "sqlite", "sqlite_db": get_data_path() / "signalblast.db", "check_same_thread": False},
    }

    get_data_path().mkdir(parents=True, exist_ok=True)

    logger = create_or_set_logger("signalblast", LOGGING_LEVEL, LOG_FILE_PATH)

    bot = BroadcasBot(config)
    await bot.load_data(
        logger=logger,
        admin_pass=admin_pass,
        expiration_time=expiration_time,
        welcome_message=welcome_message,
        instructions_url=instructions_url,
    )

    bot.register(Subscribe(bot=bot))
    bot.register(Unsubscribe(bot=bot))
    bot.register(Broadcast(bot=bot))
    bot.register(DisplayHelp(bot=bot))
    bot.register(AddAdmin(bot=bot))
    bot.register(RemoveAdmin(bot=bot))
    bot.register(BanSubscriber(bot=bot))
    bot.register(LiftBanSubscriber(bot=bot))
    bot.register(SetPing(bot=bot), contacts=False, groups=True)
    bot.register(UnsetPing(bot=bot), contacts=False, groups=True)
    bot.register(MessageToAdmin(bot=bot))
    bot.register(MessageFromAdmin(bot=bot))
    bot.register(LastMsgUserUuid(bot=bot))

    bot.scheduler.add_job(bot.delete_old_timestamps, "interval", days=1)

    if health_check_receiver is not None:
        bot.health_check_task = asyncio.create_task(health_check(bot, health_check_receiver, health_check_port))

    bot.log_rollover_task = asyncio.create_task(rotate_logs_periodically(bot))

    return bot


if __name__ == "__main__":
    four_weeks = 60 * 60 * 24 * 7 * 4  # Number of seconds in 4 weeks

    args_parser = argparse.ArgumentParser()
    args_parser.add_argument(
        "--admin_pass",
        type=str,
        help="the password to add or remove admins",
        default=os.environ.get("SIGNALBLAST_PASSWORD"),
    )
    args_parser.add_argument(
        "--expiration_time",
        type=int,
        default=os.environ.get("SIGNALBLAST_EXPIRATION_TIME", four_weeks),
        help="the expiration time for the chats in seconds",
    )

    args_parser.add_argument(
        "--signal_service",
        type=str,
        default=os.environ.get("SIGNAL_SERVICE", "localhost:8080"),
        help="the address of the signal cli rest api",
    )

    args_parser.add_argument(
        "--phone_number",
        type=str,
        default=os.environ.get("SIGNALBLAST_PHONE_NUMBER"),
        help="the phone number of the bot",
    )

    args_parser.add_argument(
        "--welcome_message",
        type=str,
        default=os.environ.get("SIGNALBLAST_WELCOME_MESSAGE"),
        help="the initial message that the user receives",
    )

    args_parser.add_argument(
        "--health_check_port",
        type=int,
        default=os.environ.get("SIGNALBLAST_HEALTHCHECK_PORT", "15556"),
        help="the port that will be listening for health checks requests",
    )

    args_parser.add_argument(
        "--health_check_receiver",
        type=str,
        default=os.environ.get("SIGNALBLAST_HEALTHCHECK_RECEIVER"),
        help="the contact or group to send messages for health checks",
    )

    args_parser.add_argument(
        "--instructions_url",
        type=str,
        default=os.environ.get("SIGNALBLAST_INSTRUCTIONS_URL"),
        help="URL for the help message",
    )

    args = args_parser.parse_args()

    if args.phone_number is None:
        value_error_msg = "The bot phone number is not set"
        raise ValueError(value_error_msg)

    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    bot = loop.run_until_complete(
        initialise_bot(
            signal_service=args.signal_service,
            phone_number=args.phone_number,
            admin_pass=args.admin_pass,
            expiration_time=args.expiration_time,
            welcome_message=args.welcome_message,
            health_check_port=args.health_check_port,
            health_check_receiver=args.health_check_receiver,
            instructions_url=args.instructions_url,
        ),
    )
    bot.start()



import asyncio
from logging.handlers import TimedRotatingFileHandler

from signalblast.broadcastbot import BroadcasBot


async def rotate_logs_periodically(bot: BroadcasBot) -> None:
    # Ensure the logs are rotated periodically even if no new log entries are made
    if len(bot.logger.handlers) == 0:
        return

    handler = bot.logger.handlers[0]

    if not isinstance(handler, TimedRotatingFileHandler):
        return

    while True:
        if handler.shouldRollover(None):
            handler.doRollover()

        await asyncio.sleep(60 * 60 * 12)  # Check every 12 hours
